/**
 * \file hydroParams.h
 * \brief Hydrodynamics solver parameters.
 *
 * \date July, 8 2013
 * \author P. Kestener
 */
#ifndef HYDRO_PARAMS_H_
#define HYDRO_PARAMS_H_

#include "real_type.h"
#include <stdbool.h> // for boolean type (C99 addon)

#include <catalyzer.h>

#define NDIM 2

//! hydro field indexes
enum ComponentIndex {
  ID=0,  /*!< ID Density field index */
  IP=1,  /*!< IP Pressure/Energy field index */
  IU=2,  /*!< X velocity / momentum index */
  IV=3  /*!< Y velocity / momentum index */ 
};

//! face index
enum FaceIdType {
  FACE_XMIN=0,
  FACE_XMAX=1,
  FACE_YMIN=2,
  FACE_YMAX=3
};

//! Riemann solver type for hydro fluxes
enum RiemannSolverType {
  APPROX, /*!< quasi-exact Riemann solver (hydro-only) */ 
  HLL,    /*!< HLL hydro and MHD Riemann solver */
  HLLC    /*!< HLLC hydro-only Riemann solver */ 
};

//! type of boundary condition (note that BC_COPY is only used in the
//! MPI version for inside boundary)
enum BoundaryConditionType {
  BC_UNDEFINED, 
  BC_DIRICHLET,   /*!< reflecting border condition */
  BC_NEUMANN,     /*!< absorbing border condition */
  BC_PERIODIC,    /*!< periodic border condition */
  BC_COPY         /*!< only used in MPI parallelized version */
};

//! enum component index
enum ComponentIndex2D {
  IX = 0,
  IY = 1
};

//! direction used in directional splitting scheme
enum Direction {
  XDIR=1, 
  YDIR=2
};

//! location of the outside boundary
enum BoundaryLocation {
  XMIN = 0, 
  XMAX = 1, 
  YMIN = 2, 
  YMAX = 3
};

//! implementation version
enum ImplementationVersion {
  IMPL_VERSION_0,
  IMPL_VERSION_1
};

// variable names in the order as in component index
extern const char * varNames[4];

/*
 * Hydro Parameters (declaration)
 */

// run parameters
extern int    nStepmax; /*!< maximun number of time steps. */
extern real_t tEnd;     /*!< end of simulation time. */
extern int    nOutput;  /*!< number of time steps between 2 consecutive outputs. */

// geometry parameters
extern int nx;     /*!< logical size along X (without ghost cells).*/
extern int ny;     /*!< logical size along Y (without ghost cells).*/
extern const int ghostWidth;  
extern int imin;   /*!< index minimum at X border*/
extern int imax;   /*!< index maximum at X border*/
extern int jmin;   /*!< index minimum at Y border*/
extern int jmax;   /*!< index maximum at Y border*/

extern int isize;  /*!< total size (in cell unit) along X direction with ghosts.*/
extern int jsize;  /*!< total size (in cell unit) along Y direction with ghosts.*/

extern real_t xmin;
extern real_t xmax;
extern real_t ymin;
extern real_t ymax;
extern real_t dx;       /*!< x resolution */
extern real_t dy;       /*!< y resolution */

extern int boundary_type_xmin;
extern int boundary_type_xmax;
extern int boundary_type_ymin;
extern int boundary_type_ymax;

// MPI parameters
extern int mx; /*!< number of MPI proc along x */
extern int my; /*!< number of MPI proc along y */

extern int iProc; /*!< x-coordinate of MPI proc on the MPI grid */
extern int jProc; /*!< y-coordinate of MPI proc on the MPI grid */

// IO parameters
extern bool ioVTK;   /*!< enable VTK  output file format (using VTI).*/
extern bool ioHDF5;  /*!< enable HDF5 output file format.*/

// hydro (numerical scheme) parameters
extern real_t gamma0;      /*!< specific heat capacity ratio (adiabatic index)*/
extern real_t gamma6;
extern real_t cfl;         /*!< Courant-Friedrich-Lewy parameter.*/
extern real_t slope_type;  /*!< type of slope computation (2 for second order scheme).*/
extern int    iorder;      /*!< */
extern real_t smallr;      /*!< small density cut-off*/
extern real_t smallc;      /*!< small speed of sound cut-off*/
extern real_t smallp;      /*!< small pressure cut-off*/
extern real_t smallpp;     /*!< smallp times smallr*/
extern int niter_riemann;  /*!< number of iteration usd in quasi-exact riemann solver*/
extern char problem[256];

// blast problem parameter
extern real_t blast_radius;
extern int blast_center_x;
extern int blast_center_y;
extern real_t blast_density_in;
extern real_t blast_density_out;
extern real_t blast_pressure_in;
extern real_t blast_pressure_out;

// other parameters
#define nbVar 4
extern Catalyzer * catalyzer;
// extern int nbVar;  /*!< number of fields in simulation (density, energy, vx, vy)*/
extern int implementationVersion; /*!< triggers which implementation to use (currently 2 versions)*/

/**
 * initialize hydrodynamics simulation run parameters.
 *
 * Read parameter from input file (INI file format).
 */
void initHydroParameters(int argc, char* argv[]);

/**
 * dump hydro parameters on stdout
 */
void printHydroParameters();

#endif // HYDRO_PARAMS_H_
