#include <string.h> // for strcmp
#include <stdio.h>
#include <stdbool.h>

#include <mpi.h>

#include "hydroRun.h"
#include "hydroParams.h"
#include "hydroUtils.h"
#include "monitoring.h"

#include <vtkSmartPointer.h>
#include <vtkCellData.h>
#include <vtkImageData.h>
#include <vtkXMLImageDataReader.h>
#include <vtkXMLImageDataWriter.h>
#include <vtkImageDataToPointSet.h>
#include <vtkImageDataGeometryFilter.h>
#include <vtkAppendFilter.h>
#include <vtkFloatArray.h>
#include <vtkDoubleArray.h>


static bool isBigEndian()
{
  const int i = 1;
  return ( (*(char*)&i) == 0 );
}

// =======================================================
// =======================================================
/**
 *
 */
void initHydroRun(HydroWorkspace* workspace)
{

  /*
   * memory allocation (use sizes with ghosts included)
   */
  workspace->U  = HydroArray2d_new(isize,jsize,nbVar);
  workspace->U2 = HydroArray2d_new(isize,jsize,nbVar);
  workspace->Q  = HydroArray2d_new(isize,jsize,nbVar);

  if (implementationVersion == 1) {
    workspace->Qm_x  = HydroArray2d_new(isize,jsize,nbVar);
    workspace->Qm_y  = HydroArray2d_new(isize,jsize,nbVar);
    workspace->Qp_x  = HydroArray2d_new(isize,jsize,nbVar);
    workspace->Qp_y  = HydroArray2d_new(isize,jsize,nbVar);
  }

  // default riemann solver
  workspace->riemann_solver_fn = riemann_approx;

  /*
   * initialize hydro array at t=0
   */
  if (strcmp(problem, "implode") == 0) {

    init_implode(workspace->U);
    init_implode(workspace->U2);

  } else if (strcmp(problem, "blast") == 0) {

    init_blast(workspace->U);
    init_blast(workspace->U2);

  } else {

    fprintf(stdout, "Problem : %s is not recognized / implemented in initHydroRun.\n",problem);
    fprintf(stdout, "Use default - implode \n");
    init_implode(workspace->U);
    init_implode(workspace->U2);

  }

} // initHydroRun


// =======================================================
// =======================================================
/**
 *
 */
void cleanupHydroRun(HydroWorkspace* workspace)
{

  HydroArray2d_free(workspace->U );
  HydroArray2d_free(workspace->U2);
  HydroArray2d_free(workspace->Q);

  if (implementationVersion == 1) {
    HydroArray2d_free(workspace->Qm_x  );
    HydroArray2d_free(workspace->Qm_y  );
    HydroArray2d_free(workspace->Qp_x  );
    HydroArray2d_free(workspace->Qp_y  );
  }

} // cleanupHydroRun

// =======================================================
// =======================================================
/**
 * Compute time step satisfying CFL condition.
 *
 * \param[in] workspace hydrodynamics data arrays
 * \param[in] useU integer, if 0 use data in U else use U2
 *
 * \return dt time step
 */
real_t compute_dt(HydroWorkspace* workspace, int useU)
{

  real_t dt;
  real_t dt_global;

  // local variables
  real_t invDt = ZERO_F;
  int    i,j;
  //real_t uLoc[nbVar]; // conservative variables in current cell
  real_t qLoc[nbVar]; // primitive    variables in current cell
  real_t c; // speed of sound
  HydroArray2d* U;
  
  // which array is the current one ?
  if (useU == 0)
    U = workspace->U;
  else
    U = workspace->U2;

  // for loop over inner region
  for (j=ghostWidth; j<=jsize-ghostWidth; j++)
    for (i=ghostWidth; i<=isize-ghostWidth; i++) {

      real_t vx, vy;
      
      // get primitive variables in current cell
      computePrimitives_ij(U, i, j, &c, qLoc);
      vx = c+FABS(qLoc[IU]);
      vy = c+FABS(qLoc[IV]);

      invDt = FMAX(invDt, vx/dx + vy/dy);

    } // end for i,j
    
  dt = cfl / invDt;

  // synchronize all mpi proc to ensure they all have computed local dt
  MPI_Barrier(MPI_COMM_WORLD);

  // perform MPI reduce
  MPI_Allreduce(&dt, &dt_global, 1, MPI_DOUBLE, MPI_MIN, MPI_COMM_WORLD);

  return dt_global;

} // compute_dt

// =======================================================
// =======================================================
// ///////////////////////////////////////////
// Wrapper to the actual computation routine
// ///////////////////////////////////////////
void godunov_unsplit(HydroWorkspace* workspace, int nStep, real_t dt)
{
  
  HydroArray2d* U;
  HydroArray2d* U2;
  U  = workspace->U;  
  U2 = workspace->U2;  

  if ( nStep % 2 == 0 ) {
    godunov_unsplit_cpu(workspace, U , U2, dt, nStep);
  } else {
    godunov_unsplit_cpu(workspace, U2, U , dt, nStep);
  }
  
} // godunov_unsplit

// =======================================================
// =======================================================
// ///////////////////////////////////////////
// Actual CPU computation of Godunov scheme
// ///////////////////////////////////////////
void godunov_unsplit_cpu(HydroWorkspace* workspace,
			 HydroArray2d* data_in, 
			 HydroArray2d* data_out, 
			 real_t dt, 
			 int nStep)
{

  // local variables
  int i, j, iVar;
  
  real_t dtdx;
  real_t dtdy;
  
  // local primitive variables
  real_t qLoc[nbVar]; // local primitive variables
  real_t* Qdata = workspace->Q->data;
  int ijSize = isize*jsize;

  // local primitive variables in neighbor cell
  real_t qLocNeighbor[nbVar];

  // local primitive variables in neighborbood
  real_t qNeighbors[4][nbVar];

  // Local slopes and neighbor slopes
  real_t dqX[nbVar];
  real_t dqY[nbVar];
  real_t dqX_neighbor[nbVar];
  real_t dqY_neighbor[nbVar];

  // rieman solver function pointer
  riemann_solver_t riemann_2d = workspace->riemann_solver_fn;

  // Local variables for Riemann problems solving
  real_t qleft[nbVar];
  real_t qright[nbVar];
  real_t qgdnv[nbVar];
  real_t flux_x[nbVar];
  real_t flux_y[nbVar];

  dtdx = dt / dx;
  dtdy = dt / dy;

  // fill ghost cell in data_in
  timerStart(&boundaries_timer);
  make_boundaries_internal(data_in);
  make_boundaries_external(data_in);
  MPI_Barrier(MPI_COMM_WORLD);
  timerStop(&boundaries_timer);
    
  // copy data_in into data_out (not necessary)
  // data_out = data_in;
  memcpy(data_out->data, data_in->data, isize * jsize * nbVar * sizeof(real_t));

  // start main computation
  timerStart(&godunov_timer);

  // convert conservative variable into primitives ones for the entire domain
  convertToPrimitives(data_in, workspace->Q);
  
  if (implementationVersion == 0) {
    
    //$omp_tmp parallel default(shared) private(qm_x,qm_y,qp_x,qp_y)
    //$omp_tmp for collapse(2) schedule(auto)
    for ( j=ghostWidth; j<=jsize-ghostWidth; j++ ) {
      for ( i=ghostWidth; i<=isize-ghostWidth; i++ ) {
	
	//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
	// deal with left interface along X !
	//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
	int indexBase = i  + isize * j;
      
	// get primitive variables state vector
	for ( iVar=0; iVar<nbVar; iVar++ ) {
	  int index = indexBase + ijSize*iVar;

	  qLoc[iVar]          = Qdata[index];
	  qNeighbors[0][iVar] = Qdata[index+1];
	  qNeighbors[1][iVar] = Qdata[index-1];
	  qNeighbors[2][iVar] = Qdata[index+isize];
	  qNeighbors[3][iVar] = Qdata[index-isize];

	} // end for iVar

	slope_unsplit_hydro_2d(qLoc, 
			       qNeighbors[0], qNeighbors[1], 
			       qNeighbors[2], qNeighbors[3],
			       dqX, dqY);
	
	// slopes at left neighbor along X
	for ( iVar=0; iVar<nbVar; iVar++ ) {
	  int index = indexBase + ijSize*iVar;

	  qLocNeighbor[iVar]  = Qdata[index-1];
	  qNeighbors[0][iVar] = Qdata[index];
	  qNeighbors[1][iVar] = Qdata[index-2];
	  qNeighbors[2][iVar] = Qdata[index-1+isize];
	  qNeighbors[3][iVar] = Qdata[index-1-isize];

	} // end for iVar
	
	slope_unsplit_hydro_2d(qLocNeighbor, 
			       qNeighbors[0], qNeighbors[1], 
			       qNeighbors[2], qNeighbors[3],
			       dqX_neighbor, dqY_neighbor);

	//
	// compute reconstructed states at left interface along X
	//
	
	// left interface : right state
	trace_unsplit_2d_along_dir(qLoc,
				   dqX, dqY,
				   dtdx, dtdy, FACE_XMIN, qright);

	// left interface : left state
	trace_unsplit_2d_along_dir(qLocNeighbor,
				   dqX_neighbor,dqY_neighbor,
				   dtdx, dtdy, FACE_XMAX, qleft);

	// Solve Riemann problem at X-interfaces and compute X-fluxes
	riemann_2d(qleft,qright,qgdnv,flux_x);
	
	//
	// update hydro array
	//
	HydroArray2d_incr (data_out,i-1,j  ,ID,- flux_x[ID]*dtdx);
	HydroArray2d_incr (data_out,i-1,j  ,IP,- flux_x[IP]*dtdx);
	HydroArray2d_incr (data_out,i-1,j  ,IU,- flux_x[IU]*dtdx);
	HydroArray2d_incr (data_out,i-1,j  ,IV,- flux_x[IV]*dtdx);

	HydroArray2d_incr (data_out,i  ,j  ,ID,  flux_x[ID]*dtdx);
	HydroArray2d_incr (data_out,i  ,j  ,IP,  flux_x[IP]*dtdx);
	HydroArray2d_incr (data_out,i  ,j  ,IU,  flux_x[IU]*dtdx);
	HydroArray2d_incr (data_out,i  ,j  ,IV,  flux_x[IV]*dtdx);

	//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
	// deal with left interface along Y !
	//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

	// slopes at left neighbor along Y
	for ( iVar=0; iVar<nbVar; iVar++ ) {
	  int index = indexBase + ijSize*iVar;

	  qLocNeighbor[iVar]  = Qdata[index-isize];
	  qNeighbors[0][iVar] = Qdata[index+1-isize];
	  qNeighbors[1][iVar] = Qdata[index-1-isize];
	  qNeighbors[2][iVar] = Qdata[index];
	  qNeighbors[3][iVar] = Qdata[index-2*isize];

	} // end for iVar
	
	slope_unsplit_hydro_2d(qLocNeighbor, 
			       qNeighbors[0], qNeighbors[1], 
			       qNeighbors[2], qNeighbors[3],
			       dqX_neighbor, dqY_neighbor);
	//
	// compute reconstructed states at left interface along Y
	//
	
	// left interface : right state
	trace_unsplit_2d_along_dir(qLoc,
				   dqX, dqY,
				   dtdx, dtdy, FACE_YMIN, qright);

	// left interface : left state
	trace_unsplit_2d_along_dir(qLocNeighbor,
				   dqX_neighbor,dqY_neighbor,
				   dtdx, dtdy, FACE_YMAX, qleft);

	// Solve Riemann problem at Y-interfaces and compute Y-fluxes
	swapValues(&(qleft[IU]) ,&(qleft[IV]) );
	swapValues(&(qright[IU]),&(qright[IV]));
	riemann_2d(qleft,qright,qgdnv,flux_y);

	//
	// update hydro array
	//
	HydroArray2d_incr (data_out,i  ,j-1,ID,- flux_y[ID]*dtdy);
	HydroArray2d_incr (data_out,i  ,j-1,IP,- flux_y[IP]*dtdy);
	HydroArray2d_incr (data_out,i  ,j-1,IU,- flux_y[IV]*dtdy); // IU/IV swapped
	HydroArray2d_incr (data_out,i  ,j-1,IV,- flux_y[IU]*dtdy); // IU/IV swapped

	HydroArray2d_incr (data_out,i  ,j  ,ID,  flux_y[ID]*dtdy);
	HydroArray2d_incr (data_out,i  ,j  ,IP,  flux_y[IP]*dtdy);
	HydroArray2d_incr (data_out,i  ,j  ,IU,  flux_y[IV]*dtdy); // IU/IV swapped
	HydroArray2d_incr (data_out,i  ,j  ,IV,  flux_y[IU]*dtdy); // IU/IV swapped

      } // end for j
    } // end for i
    
  } else if (implementationVersion == 1) {
    
    // trace computation: fill arrays qm_x, qm_y, qp_x, qp_y
    computeTrace(workspace, data_in, dt);

    // Compute flux via Riemann solver and update (time integration)
    computeFluxesAndUpdate(workspace, data_out, dt);

  } // end implementationVersion == 1
  
  timerStop(&godunov_timer);
  
} // godunov_unsplit_cpu

// =======================================================
// =======================================================
// ///////////////////////////////////////////////////////////////////
// Convert conservative variables array U into primitive var array Q
// ///////////////////////////////////////////////////////////////////
void convertToPrimitives(HydroArray2d *U, HydroArray2d *Q)
{

  // local variables
  // primitive variable state vector
  real_t qLoc[nbVar];
  real_t c;
  int    i,j;
    
  real_t* Qdata = Q->data;
  int isize = Q->isize;
  int ijSize = Q->isize*Q->jsize;

  for ( j=0; j<jsize; j++) {
    for ( i=0; i<isize; i++) {
      
      computePrimitives_ij(U, i, j, &c, qLoc);
      
      // copy q state in q global
      Qdata[i+isize*j+ijSize*ID] = qLoc[ID];
      Qdata[i+isize*j+ijSize*IP] = qLoc[IP];
      Qdata[i+isize*j+ijSize*IU] = qLoc[IU];
      Qdata[i+isize*j+ijSize*IV] = qLoc[IV];

    } // end for i
  } // end for j
  
} // convertToPrimitives

// =======================================================
// =======================================================
// ///////////////////////////////////////////////////////////////////
// Compute trace (only used in implementation version 1), i.e.
// fill global array qm_x, qmy, qp_x, qp_y
// ///////////////////////////////////////////////////////////////////
void computeTrace(HydroWorkspace* workspace, HydroArray2d *U, real_t dt)
{

  // local variables
  int i,j,iVar;
  real_t qLoc   [nbVar]; // local primitive variables
  real_t qPlusX [nbVar];
  real_t qMinusX[nbVar];
  real_t qPlusY [nbVar];
  real_t qMinusY[nbVar];
  real_t dq[2][nbVar];
  real_t qm[2][nbVar], qp[2][nbVar];
  real_t dtdx;
  real_t dtdy;
  real_t* Q    = workspace->Q->data;
  real_t* Qm_x = workspace->Qm_x->data;
  real_t* Qm_y = workspace->Qm_y->data;
  real_t* Qp_x = workspace->Qp_x->data;
  real_t* Qp_y = workspace->Qp_y->data;

  int isize = workspace->Q->isize;
  int ijSize = 
    (workspace->Q->isize)*
    (workspace->Q->jsize);

  dtdx = dt / dx;
  dtdy = dt / dy;

  for ( j=1; j<=jsize-2; j++) {
    for ( i=1; i<=isize-2; i++) {

      int indexBase = i  + isize * j;
      
      // get primitive variables state vector
      for ( iVar=0; iVar<nbVar; iVar++ ) {
	int index = indexBase + ijSize*iVar;
	qLoc   [iVar] = Q[index];
	qPlusX [iVar] = Q[index+1];
	qMinusX[iVar] = Q[index-1];
	qPlusY [iVar] = Q[index+isize];
	qMinusY[iVar] = Q[index-isize];
      } // end for iVar
      
      // get hydro slopes dq
      slope_unsplit_hydro_2d(qLoc, qPlusX, qMinusX, qPlusY, qMinusY, dq[0], dq[1]);
      
      // compute qm, qp
      trace_unsplit_hydro_2d(qLoc, dq, dtdx, dtdy, qm, qp);

      // store qm, qp : only what is really needed
      for ( iVar=0; iVar<nbVar; iVar++ ) {
	int index = indexBase + ijSize*iVar;
	Qm_x[index] = qm[0][iVar];
	Qp_x[index] = qp[0][iVar];
	Qm_y[index] = qm[1][iVar];
	Qp_y[index] = qp[1][iVar];
      } // end for iVar

    } // end for i
  } // end for j

} // computeTrace

// =======================================================
// =======================================================
// //////////////////////////////////////////////////////////////////
// Compute flux via Riemann solver and update (time integration)
// //////////////////////////////////////////////////////////////////
void computeFluxesAndUpdate(HydroWorkspace* workspace, 
			    HydroArray2d *U, 
			    real_t dt)
{
 
  // local variables
  int i,j;
  real_t qleft[nbVar], qright[nbVar];
  real_t flux_x[nbVar], flux_y[nbVar];
  real_t qgdnv[nbVar];
  
  real_t dtdx = dt / dx;
  real_t dtdy = dt / dy;


  real_t* Udata = U->data;
  real_t* Qm_x = workspace->Qm_x->data;
  real_t* Qm_y = workspace->Qm_y->data;
  real_t* Qp_x = workspace->Qp_x->data;
  real_t* Qp_y = workspace->Qp_y->data;

  int isize = U->isize;
  int ijSize = U->isize * U->jsize;

  riemann_solver_t riemann_2d = workspace->riemann_solver_fn;

  for ( j=ghostWidth; j<=jsize-ghostWidth; j++ ) {
    for ( i=ghostWidth; i<=isize-ghostWidth; i++ ) {

      int index;

      //
      // Solve Riemann problem at X-interfaces and compute
      // X-fluxes
      //
      index = i-1 + isize * j;
      qleft[ID]   = Qm_x[index + ijSize * ID];
      qleft[IP]   = Qm_x[index + ijSize * IP];
      qleft[IU]   = Qm_x[index + ijSize * IU];
      qleft[IV]   = Qm_x[index + ijSize * IV];
      
      index = i + isize * j;
      qright[ID]  = Qp_x[index + ijSize * ID];
      qright[IP]  = Qp_x[index + ijSize * IP];
      qright[IU]  = Qp_x[index + ijSize * IU];
      qright[IV]  = Qp_x[index + ijSize * IV];
      
      // compute hydro flux_x
      riemann_2d(qleft,qright,qgdnv,flux_x);

      //
      // Solve Riemann problem at Y-interfaces and compute Y-fluxes
      //
      index = i + isize * (j-1);
      qleft[ID]   = Qm_y[index + ijSize * ID];
      qleft[IP]   = Qm_y[index + ijSize * IP];
      qleft[IU]   = Qm_y[index + ijSize * IV]; // watchout IU, IV permutation
      qleft[IV]   = Qm_y[index + ijSize * IU]; // watchout IU, IV permutation

      index = i + isize * j;
      qright[ID]  = Qp_y[index + ijSize * ID];
      qright[IP]  = Qp_y[index + ijSize * IP];
      qright[IU]  = Qp_y[index + ijSize * IV]; // watchout IU, IV permutation
      qright[IV]  = Qp_y[index + ijSize * IU]; // watchout IU, IV permutation
      
      // compute hydro flux_y
      riemann_2d(qleft,qright,qgdnv,flux_y);
            
      //
      // update hydro array
      //
      index = i-1 + isize*j;
      Udata[index + ijSize * ID] += - flux_x[ID]*dtdx;
      Udata[index + ijSize * IP] += - flux_x[IP]*dtdx;
      Udata[index + ijSize * IU] += - flux_x[IU]*dtdx;
      Udata[index + ijSize * IV] += - flux_x[IV]*dtdx;

      index = i + isize*j;
      Udata[index + ijSize * ID] +=   flux_x[ID]*dtdx;
      Udata[index + ijSize * IP] +=   flux_x[IP]*dtdx;
      Udata[index + ijSize * IU] +=   flux_x[IU]*dtdx;
      Udata[index + ijSize * IV] +=   flux_x[IV]*dtdx;

      index = i + isize*(j-1);
      Udata[index + ijSize * ID] += - flux_y[ID]*dtdy;
      Udata[index + ijSize * IP] += - flux_y[IP]*dtdy;
      Udata[index + ijSize * IU] += - flux_y[IV]*dtdy; // watchout IU and IV swapped
      Udata[index + ijSize * IV] += - flux_y[IU]*dtdy; // watchout IU and IV swapped

      index = i + isize*j;
      Udata[index + ijSize * ID] +=   flux_y[ID]*dtdy;
      Udata[index + ijSize * IP] +=   flux_y[IP]*dtdy;
      Udata[index + ijSize * IU] +=   flux_y[IV]*dtdy; // watchout IU and IV swapped
      Udata[index + ijSize * IV] +=   flux_y[IU]*dtdy; // watchout IU and IV swapped

    } // end for i
  } // end for j

 
} // computeFluxesAndUpdate

// =======================================================
// =======================================================
/**
 * Hydrodynamical Implosion Test.
 * http://www.astro.princeton.edu/~jstone/Athena/tests/implode/Implode.html
 */
void init_implode(HydroArray2d* U)
{

  int isize = U->isize;
  int ijSize = U->isize * U->jsize;
  real_t* Udata = U->data;

  for (int j=0; j<jsize; j++)
    for (int i=0; i<isize; i++) {
      
      // cartesian coordinate of current cell in global grid
      int iGlob = i + iProc*nx;
      int jGlob = j + jProc*ny;

      int index = i + isize * j;

      real_t tmp = 1.0*(iGlob-ghostWidth)/(nx*mx) + 1.0*(jGlob-ghostWidth)/(ny*my);
      if (tmp > 0.5) {
	Udata[index + ID*ijSize] = 1.0;
	Udata[index + IP*ijSize] = 1.0/(gamma0-1.0);
	Udata[index + IU*ijSize] = 0.0;
	Udata[index + IV*ijSize] = 0.0;
      } else {
	Udata[index + ID*ijSize] = 0.125;
	Udata[index + IP*ijSize] = 0.14/(gamma0-1.0);
	Udata[index + IU*ijSize] = 0.0;
	Udata[index + IV*ijSize] = 0.0;
      }
    }

} // init_implode

// =======================================================
// =======================================================
/**
 * Hydrodynamical blast Test.
 * http://www.astro.princeton.edu/~jstone/Athena/tests/blast/blast.html
 */
void init_blast(HydroArray2d* U)
{

  int isize = U->isize;
  int ijSize = U->isize * U->jsize;
  real_t* Udata = U->data;

  real_t radius2 = blast_radius*blast_radius;
  
  for (int j=0; j<jsize; j++)
    for (int i=0; i<isize; i++) {

      
      // cartesian coordinate of current cell in global grid
      int iGlob = i + iProc*nx;
      int jGlob = j + jProc*ny;

      int index = i + isize * j;

      real_t d2 = 
	(iGlob-blast_center_x)*(iGlob-blast_center_x)+
	(jGlob-blast_center_y)*(jGlob-blast_center_y);

      if (d2 < radius2) {
	Udata[index + ID*ijSize] = blast_density_in;
	Udata[index + IP*ijSize] = blast_pressure_in/(gamma0-1.0);
	Udata[index + IU*ijSize] = 0.0;
	Udata[index + IV*ijSize] = 0.0;
      } else {
	Udata[index + ID*ijSize] = blast_density_out;
	Udata[index + IP*ijSize] = blast_pressure_out/(gamma0-1.0);
	Udata[index + IU*ijSize] = 0.0;
	Udata[index + IV*ijSize] = 0.0;
      }
    }

} // init_blast

// =======================================================
// =======================================================
void write_pvtk_wrapper(int iStep)
{

  // local variables
  FILE *fp;
  char pvtk_filename[80] = "";
  char stepNum[8];
  int iVar, iPiece;

  // check scalar data type
  bool useDouble = false;

  // mpi parameters
  int  myRank, nProc;


  
  MPI_Comm_size(MPI_COMM_WORLD, &nProc);
  MPI_Comm_rank(MPI_COMM_WORLD, &myRank);

  if (sizeof(real_t) == sizeof(double)) {
    useDouble = true;
  }
    
  // write iStep in string stepNum
  sprintf(stepNum,"%07d",iStep);

  // open pvtk file 
  strcat(pvtk_filename, "euler2d_mpi");
  strcat(pvtk_filename, "_step_");
  strcat(pvtk_filename, stepNum);
  strcat(pvtk_filename, ".pvti");
  fp = fopen(pvtk_filename, "w");

  // write pvtk header
  fprintf(fp, "<?xml version=\"1.0\"?>\n");
  if (isBigEndian())
    fprintf(fp,"<VTKFile type=\"PImageData\" version=\"0.1\" byte_order=\"BigEndian\">\n");
  else
    fprintf(fp,"<VTKFile type=\"PImageData\" version=\"0.1\" byte_order=\"LittleEndian\">\n");
  fprintf(fp,"  <PImageData WholeExtent=\"%d %d %d %d %d %d\" GhostLevel=\"0\" Origin=\"0 0 0\" Spacing=\"1 1 1\">\n",0,mx*nx,0,my*ny,0,1);
  fprintf(fp,"    <PCellData Scalars=\"Scalars_\">\n");
  for (iVar=0; iVar<nbVar; iVar++) {
    if (useDouble)
      fprintf(fp,"      <PDataArray type=\"Float64\" Name=\"%s\"/>\n", varNames[iVar]);
    else
      fprintf(fp,"      <PDataArray type=\"Float32\" Name=\"%s\"/>\n", varNames[iVar]);
  }
  fprintf(fp,"    </PCellData>\n");
  
  // write pvtk data
  // one piece per MPI process (identified by iPiece)
  for (iPiece=0; iPiece<nProc; ++iPiece) {
    char mpiNum[8];
    char filename[256] = "";
    int iProc, jProc;
    
    // write ipiece (mpi rank) in string mpiNum
    sprintf(mpiNum,"%07d",iPiece);
    
    // concatenate file prefix + file number + suffix
    strcat(filename, "euler2d_mpi_");
    strcat(filename, mpiNum);
    strcat(filename, "_step_");
    strcat(filename, stepNum);
    strcat(filename, ".vti");

    jProc = iPiece / mx;
    iProc = iPiece - mx*jProc;
    fprintf(fp, "    <Piece Extent=\"%d %d %d %d %d %d\" Source=\"%s\"/>\n",
  	    iProc*nx,iProc*nx+nx,jProc*ny,jProc*ny+ny,0,1,filename);
  } // end for iPiece

  // write footer
  fprintf(fp, "</PImageData>\n");
  fprintf(fp, "</VTKFile>\n");
 
  // close file
  fclose(fp);

} // write_pvtk_wrapper

// =======================================================
// =======================================================
// ///////////////////////////////////////////////////////
// output routine (VTK file format, ASCII, VtkImageData)
// ///////////////////////////////////////////////////////
void saveVTK(HydroArray2d* U, int iStep)
{
  
  // local variables
  int i,j,iVar;
  char filename[80] = "";
  char stepNum[8];
  char mpiNum[8];
  FILE *fp;

  // check scalar data type
  bool useDouble = false;

  real_t* Udata = U->data;
  int isize = U->isize;
  int ijSize = U->isize * U->jsize;

  // mpi parameters
  int  myRank, nProc;
  MPI_Comm_size(MPI_COMM_WORLD, &nProc);
  MPI_Comm_rank(MPI_COMM_WORLD, &myRank);

  if (sizeof(real_t) == sizeof(double)) {
    useDouble = true;
  }

  // rank 0 must write pvtk file wrapper
  /* if (myRank==0) */
  /*   write_pvtk_wrapper(iStep); */
  
  // write iStep in string stepNum
  sprintf(stepNum,"%07d",iStep);

  // write myRank in string mpiNum
  sprintf(mpiNum,"%07d",myRank);

  // concatenate file prefix + file number + suffix
  strcat(filename, "euler2d_mpi_");
  strcat(filename, mpiNum);
  strcat(filename, "_step_");
  strcat(filename, stepNum);
  strcat(filename, ".vti");
  
  /* VTK code for exporting or saving data */
  vtkSmartPointer<vtkImageData> imageData = vtkSmartPointer<vtkImageData>::New();
  imageData->SetOrigin(0.0,0.0,0.0);
  imageData->SetSpacing(1.0,1.0,1.0); 
  imageData->SetExtent(iProc*nx,iProc*nx+nx,jProc*ny,jProc*ny+ny,0,1);

  int* dims = imageData->GetDimensions();
  // int dims[3]; // can't do this

  printf("Dims: x: %d y: %d z: %d\n", dims[0], dims[1], dims[2]);

  printf("Number of points: %d\n", imageData->GetNumberOfPoints());
  printf("Number of cells: %d\n", imageData->GetNumberOfCells());

  for ( iVar=0; iVar<nbVar; iVar++) {
      vtkSmartPointer<vtkDataArray> dataArray;
      if (useDouble)
      { dataArray = vtkSmartPointer<vtkDoubleArray>::New(); }
      else
      { dataArray = vtkSmartPointer<vtkFloatArray>::New(); }
      dataArray->SetName(varNames[iVar]);
      dataArray->SetNumberOfComponents(1);
      dataArray->SetNumberOfTuples(nx * ny);
      //dataArray->Allocate(nx * ny);

      int k = 0;
      for ( j=jmin+ghostWidth; j<=jmax-ghostWidth; j++) {
          for ( i=imin+ghostWidth; i<=imax-ghostWidth; i++) {
              if (useDouble)
              {
                  dataArray->SetTuple1(k, Udata[i + isize * j + ijSize * iVar]);
              }
              else
              { 
                  dataArray->SetTuple1(k, Udata[i + isize * j + ijSize * iVar]);
              }
              k++;
          }
      }
      imageData->GetCellData()->AddArray(dataArray);
  } // end for iVar

  /*vtkSmartPointer<vtkXMLImageDataWriter> writer = vtkSmartPointer<vtkXMLImageDataWriter>::New();*/
  /*writer->SetInputData(imageData);*/
  /*writer->SetFileName(filename);*/
  /*writer->Update();*/

  catalyzer->Save(iStep, (double)iStep, imageData);

} // saveVTK

// =======================================================
// =======================================================
// //////////////////////////////////////////////////
// Fill ghost cells according to border condition :
// absorbant, reflexive (no MPI comm required) 
// periodic border conditions are done by default by
// make_boundaries_internal
// //////////////////////////////////////////////////
void make_boundaries_external(HydroArray2d *U)
{

  // local variables
  int i,j,i0,j0,iVar;
  real_t sign;

  real_t* Udata = U->data;
  int isize = U->isize;
  int ijSize = U->isize * U->jsize;
    
  if (boundary_type_xmin == BC_PERIODIC || boundary_type_xmax == BC_PERIODIC ||
      boundary_type_ymin == BC_PERIODIC || boundary_type_ymax == BC_PERIODIC   ) {
    // don't do anything, already done in make_boundaries_internal
    return;
  }

  // boundary xmin
  if (iProc == 0) {
    for ( iVar=0; iVar<nbVar; iVar++ ) {
      for ( i=0; i<ghostWidth; i++ ) {
	sign=ONE_F;
	if ( boundary_type_xmin == BC_DIRICHLET ) {
	  i0=2*ghostWidth-1-i;
	  if (iVar==IU) sign=-ONE_F;
	} else if( boundary_type_xmin == BC_NEUMANN ) {
	  i0=ghostWidth;
	} else { // periodic
	  i0=nx+i;
	}
	
	for ( j=jmin+ghostWidth; j<=jmax-ghostWidth; j++ ) {
	  Udata[i + isize*j + ijSize*iVar] = Udata[i0 + isize*j + ijSize*iVar]*sign;
	} // end for j
      } // end for i
    } // end for iVar
  } // end iProc == 0

  // boundary xmax
  if (iProc == mx-1) {
    for ( iVar=0; iVar<nbVar; iVar++ ) {
      for ( i=nx+ghostWidth; i<=nx+2*ghostWidth-1; i++ ) {
	sign=ONE_F;
	if ( boundary_type_xmax == BC_DIRICHLET ) {
	  i0=2*nx+2*ghostWidth-1-i;
	  if (iVar==IU) sign=-ONE_F;
	} else if ( boundary_type_xmax == BC_NEUMANN ) {
	  i0=nx+ghostWidth-1;
	} else { // periodic
	  i0=i-nx;
	}
	for ( j=jmin+ghostWidth; j<=jmax-ghostWidth; j++) {
	  Udata[i + isize*j + ijSize*iVar] = Udata[i0 + isize*j + ijSize*iVar]*sign;
	} // end for j
      } // end for i
    } // end for iVar
  } // end iProc == mx-1

  // boundary ymin
  if (jProc == 0) {
    for ( iVar=0; iVar<nbVar; iVar++) {
      for ( j=0; j<ghostWidth; j++ ) {
	sign=ONE_F;
	if ( boundary_type_ymin == BC_DIRICHLET ) {
	  j0=2*ghostWidth-1-j;
	  if (iVar==IV) sign=-ONE_F;
	} else if ( boundary_type_ymin == BC_NEUMANN ) {
	  j0=ghostWidth;
	} else { // periodic
	  j0=ny+j;
	}
	
	for ( i=imin+ghostWidth; i<=imax-ghostWidth; i++ ) {
	  Udata[i + isize*j + ijSize*iVar] = Udata[i + isize*j0 + ijSize*iVar]*sign;
	} // end for i
      } // end for j
    } // end for iVar
  } // end jProc == 0

  // boundary ymax
  if (jProc == my-1) {
    for ( iVar=0; iVar<nbVar; iVar++ ) {
      for ( j=ny+ghostWidth; j<=ny+2*ghostWidth-1; j++ ) {
	sign=ONE_F;
	if ( boundary_type_ymax == BC_DIRICHLET ) {
	  j0=2*ny+2*ghostWidth-1-j;
	  if (iVar==IV) sign=-ONE_F;
	} else if ( boundary_type_ymax == BC_NEUMANN ) {
	  j0=ny+ghostWidth-1;
	} else { // periodic
	  j0=j-ny;
	}
	
	for ( i=imin+ghostWidth; i<=imax-ghostWidth; i++ ) {
	  Udata[i + isize*j + ijSize*iVar] = Udata[i + isize*j0 + ijSize*iVar]*sign;
	} // end for i
      } //end for j
    } // end for iVar
  } // end jProc == my-1

} // make_boundaries_external

// =======================================================
// =======================================================
// ////////////////////////////////////////////////////////
// Fill ghost cells using MPI communication with neighbors
// ////////////////////////////////////////////////////////
void make_boundaries_internal(HydroArray2d *U)
{

  // local variables
  int i,j,iVar;
  int msgSize; // size in bytes of border buffer
  MPI_Datatype mpiDataType;
  int neighborRankSend=0;
  int neighborRankRecv=0;
  int sendTag, recvTag;

  int iProcp1, iProcm1; /** cartesian coordinate of neighboring MPI proc */
  int jProcp1, jProcm1; /** cartesian coordinate of neighboring MPI proc */

  iProcp1 = iProc==mx-1 ? 0    : iProc+1;
  iProcm1 = iProc==0    ? mx-1 : iProc-1;

  jProcp1 = jProc==my-1 ? 0    : jProc+1;
  jProcm1 = jProc==0    ? my-1 : jProc-1;


  if (sizeof(real_t) == sizeof(float)) {
    mpiDataType = MPI_FLOAT;
  } else {
    mpiDataType = MPI_DOUBLE;
  }

  // allocate memory for border buffer
  HydroArray2d *sendBorderBufX = HydroArray2d_new(ghostWidth, jsize, nbVar);
  HydroArray2d *recvBorderBufX = HydroArray2d_new(ghostWidth, jsize, nbVar);
  HydroArray2d *sendBorderBufY = HydroArray2d_new(isize, ghostWidth, nbVar);
  HydroArray2d *recvBorderBufY = HydroArray2d_new(isize, ghostWidth, nbVar);

  /*
   * boundary XMIN / XMAX backward
   */
  // copy border XMIN
  for ( iVar=0; iVar<nbVar; iVar++ ) {
    for (j=0; j<jsize; j++) {
      for (i=0; i<ghostWidth; i++) {
	int offset = ghostWidth;
	HydroArray2d_set(sendBorderBufX,i,j,iVar, HydroArray2d_get(U,i+offset,j,iVar));
      }
    }
  }
  
  // send to   neighbor iProc-1,jProc
  // recv from neighbor iProc+1,jProc
  msgSize = ghostWidth*jsize*nbVar;
  neighborRankSend = iProcm1 + jProc*mx;
  neighborRankRecv = iProcp1 + jProc*mx;
  sendTag = recvTag = 111;
  MPI_Sendrecv(sendBorderBufX->data, msgSize, mpiDataType, neighborRankSend, sendTag,
	       recvBorderBufX->data, msgSize, mpiDataType, neighborRankRecv, recvTag,
	       MPI_COMM_WORLD,  MPI_STATUS_IGNORE);
  
  // update border XMAX
  for ( iVar=0; iVar<nbVar; iVar++ ) {
    for (j=0; j<jsize; j++) {
      for (i=0; i<ghostWidth; i++) {
	int offset = isize-ghostWidth;
	HydroArray2d_set(U,i+offset,j,iVar, HydroArray2d_get(recvBorderBufX,i,j,iVar));
      }
    }
  }
  

  /*
   * boundary XMIN/ XMAX forward
   */
  // copy border XMAX
  for ( iVar=0; iVar<nbVar; iVar++ ) {
    for (j=0; j<jsize; j++) {
      for (i=0; i<ghostWidth; i++) {
	int offset = isize-2*ghostWidth;
	HydroArray2d_set(sendBorderBufX,i,j,iVar, HydroArray2d_get(U,i+offset,j,iVar));
      }
    }
  }
  
  // send to   neighbor iProc+1,jProc
  // recv from neighbor iProc-1,jProc
  msgSize = ghostWidth*jsize*nbVar;
  neighborRankSend = iProcp1 + jProc*mx;
  neighborRankRecv = iProcm1 + jProc*mx;
  sendTag = recvTag = 111;
  MPI_Sendrecv(sendBorderBufX->data, msgSize, mpiDataType, neighborRankSend, sendTag,
	       recvBorderBufX->data, msgSize, mpiDataType, neighborRankRecv, recvTag,
	       MPI_COMM_WORLD,  MPI_STATUS_IGNORE);
  
  // update border XMIN
  for ( iVar=0; iVar<nbVar; iVar++ ) {
    for (j=0; j<jsize; j++) {
      for (i=0; i<ghostWidth; i++) {
	int offset = 0;
	HydroArray2d_set(U,i+offset,j,iVar, HydroArray2d_get(recvBorderBufX,i,j,iVar));
      }
    }
  }
  
  MPI_Barrier(MPI_COMM_WORLD);

  /*
   * boundary YMIN / YMAX backward
   */
  // copy border YMIN
  for ( iVar=0; iVar<nbVar; iVar++ ) {
    for (j=0; j<ghostWidth; j++) {
      for (i=0; i<isize; i++) {
  	int offset = ghostWidth;
  	HydroArray2d_set(sendBorderBufY,i,j,iVar, HydroArray2d_get(U,i,j+offset,iVar));
      }
    }
  }
  
  // send to   neighbor iProc,jProc-1
  // recv from neighbor iProc,jProc+1
  msgSize = isize*ghostWidth*nbVar;
  neighborRankSend = iProc + jProcm1*mx;
  neighborRankRecv = iProc + jProcp1*mx;
  sendTag = recvTag = 211;
  MPI_Sendrecv(sendBorderBufY->data, msgSize, mpiDataType, neighborRankSend, sendTag,
  	       recvBorderBufY->data, msgSize, mpiDataType, neighborRankRecv, recvTag,
  	       MPI_COMM_WORLD,  MPI_STATUS_IGNORE);
  
  // update border YMAX
  for ( iVar=0; iVar<nbVar; iVar++ ) {
    for (j=0; j<ghostWidth; j++) {
      for (i=0; i<isize; i++) {
  	int offset = jsize-ghostWidth;
  	HydroArray2d_set(U,i,j+offset,iVar, HydroArray2d_get(recvBorderBufY,i,j,iVar));
      }
    }
  }
  
  /*
   * boundary YMIN / YMAX forward
   */
  // copy border YMAX
  for ( iVar=0; iVar<nbVar; iVar++ ) {
    for (j=0; j<ghostWidth; j++) {
      for (i=0; i<isize; i++) {
  	int offset = jsize-2*ghostWidth;
  	HydroArray2d_set(sendBorderBufY,i,j,iVar, HydroArray2d_get(U,i,j+offset,iVar));
      }
    }
  }
  
  // send to   neighbor iProc,jProc+1
  // recv from neighbor iProc,jProc-1
  msgSize = isize*ghostWidth*nbVar;
  neighborRankSend = iProc + jProcp1*mx;
  neighborRankRecv = iProc + jProcm1*mx;
  sendTag = recvTag = 211;
  MPI_Sendrecv(sendBorderBufY->data, msgSize, mpiDataType, neighborRankSend, sendTag,
  	       recvBorderBufY->data, msgSize, mpiDataType, neighborRankRecv, recvTag,
  	       MPI_COMM_WORLD,  MPI_STATUS_IGNORE);
  
  // update border YMIN
  for ( iVar=0; iVar<nbVar; iVar++ ) {
    for (j=0; j<ghostWidth; j++) {
      for (i=0; i<isize; i++) {
  	int offset = 0;
  	HydroArray2d_set(U,i,j+offset,iVar, HydroArray2d_get(recvBorderBufY,i,j,iVar));
      }
    }
  }
  
  // free memory
  HydroArray2d_free(sendBorderBufX);
  HydroArray2d_free(recvBorderBufX);
  HydroArray2d_free(sendBorderBufY);
  HydroArray2d_free(recvBorderBufY);

} // make_boundaries_internal
